"""
Package for searching for images and downloading them
"""
import os
import base64
import requests
from PIL import Image
from selenium import webdriver
from selenium.webdriver.chrome.service import Service
from selenium.webdriver.chrome.options import Options
from selenium.webdriver.common.by import By
from selenium.webdriver.common.keys import Keys
from selenium.common.exceptions import NoSuchElementException
from urllib.parse import quote
import time
import logging
logging.basicConfig(level='INFO')
logger = logging.getLogger(__name__)


def __is_img_size_good(image_path, min_width, min_height):
    """
    :param image_path: string, path must exist
    :param min_width: int or None
    :param min_height: int or None
    :return: boolean, True if there are no minimum size requirements of if size requirements are satisfied
    """
    if min_width is None and min_height is None:
        return True

    try:
        with Image.open(image_path) as img:
            width, height = img.size
    except Exception as ex:
        print(f"PIL failed to open {image_path}")
        return False

    if min_width is not None:
        if width < min_width:
            return False

    if min_height is not None:
        if height < min_height:
            return False

    return True


def download_file(directory, filename, url, min_width, min_height):
    """
    :param directory: string, a valid folder that exists
    :param filename: string
    :param url: string
    :param min_width, min_height: int, if not None, then it will check the size restriction and if the file is too small
        it will delete the file and return False
    :return: False of download fails or the file is too small
        it downloads the url file into the folder with the given filename
    """
    img_data = None
    if url.startswith('data:image/'):
        try:
            img_data = base64.urlsafe_b64decode(url[url.find(',') + 1:])
        except Exception as ex:
            print(ex)
            return False

        open(os.path.join(directory, filename), "wb").write(img_data)

    else:
        try:
            response = requests.get(url)
            img_data = response.content
        except Exception as ex:
            print(ex)
            return False

    img_path = os.path.join(directory, filename)
    with open(img_path, "wb") as img_file:
        img_file.write(img_data)
        img_file.close()

    is_img_size_good = __is_img_size_good(img_path, min_width, min_height)
    if not is_img_size_good:
        os.remove(img_path)
    return is_img_size_good


class SearchEngine(object):
    """
    Base class for search engines
    """
    END_SIGNAL = '{END}'
    SCROLL_DELAY = 1.2

    def __init__(self, chrome_driver_path):
        self.chrome_driver = chrome_driver_path
        service_obj = Service(self.chrome_driver)
        options = Options()
        options.add_argument("--headless")
        options.add_argument("--log-level=3")
        options.add_argument('--no-sandbox')
        options.add_argument('--disable-dev-shm-usage')
        self.driver = webdriver.Chrome(options=options, service=service_obj)
        self.driver.set_window_size(1505, 1000)
        self.next_page_btn = None
        self.img_class_ref = None
        self.has_next_result_btn = False
        self.must_type_search = False

    def _get_search_url(self):
        """
        Interface for fetching the base url
        """
        raise NotImplementedError()

    def _get_next_btn(self):
        raise NotImplementedError()

    def _found_next_btn(self):
        self.next_page_btn = None
        try:
            next_page_btn = self._get_next_btn()
            if next_page_btn is None:
                return False
            self.next_page_btn = next_page_btn
            return next_page_btn.is_displayed()
        except NoSuchElementException:
            return False

    def __exit__(self, exc_type, exc_value, traceback):
        self.driver.quit()

    def _found_next_btn(self):
        self.next_page_btn = None
        try:
            next_page_btn = self.driver.find_element(By.CLASS_NAME, "pagination__num--next")
            self.next_page_btn = next_page_btn
            return next_page_btn.is_displayed()
        except NoSuchElementException:
            return False

    def _go_to_next_page(self):
        if self.next_page_btn is None:
            raise Exception("The next page button is not found")
        self.next_page_btn.click()
        self.next_page_btn = None
        time.sleep(1)

    def _get_search_field(self):
        raise NotImplementedError()

    def _get_image(self, search):
        """
        :param search: string keywords to search
        :return: either an image url or base64 encrypted image data
        This is a generator function
        """
        logger.info('Scraping page links...')
        search_url = self._get_search_url(search)
        self.driver.get(search_url)

        if self.must_type_search:
            search_field = self._get_search_field()
            search = quote(search.encode('utf-8'))
            search_field.send_keys(f"{search}\n")

        images = set()
        for page_num in range(10):
            body = self.driver.find_element(By.TAG_NAME, 'body')
            reached_page_end = False
            last_height = self.driver.execute_script("return document.body.scrollHeight")
            while not reached_page_end:
                body.send_keys(Keys.END)
                time.sleep(self.SCROLL_DELAY)
                new_height = self.driver.execute_script("return document.body.scrollHeight")
                if last_height == new_height:
                    reached_page_end = True
                else:
                    last_height = new_height

            for img in self.driver.find_elements(By.TAG_NAME, 'img'):
                src = img.get_attribute("src")
                class_name = img.get_attribute("class")
                class_check = True
                if self.img_class_ref is not None:
                    class_check = self.img_class_ref in class_name

                if src is not None and not src.endswith('.svg') and not src.startswith('data:image/svg') \
                        and class_check is True:
                    if src not in images:
                        images.add(src)
                        yield src

            if not self.has_next_result_btn:
                yield self.END_SIGNAL

            if self._found_next_btn():
                last_height = self.driver.execute_script("return document.body.scrollHeight")
                self._go_to_next_page()
                new_height = self.driver.execute_script("return document.body.scrollHeight")
                if new_height == last_height:
                    yield self.END_SIGNAL


class Scraper(object):
    """
    Use this class to search the internet for images and download them
    """
    def __init__(self, search_engine: SearchEngine):
        self.search_engine = search_engine

    def download(self, keywords, directory, limit=10, min_width=None, min_height=None,  prefix='', postfix=''):
        """
        :param keywords: string
        :param directory: string, valid path
        :param limit: int
        :param min_width: int
        :param min_height: int
        :param prefix: string
        :param postfix: string
        :return: Nothing. It downloads files in the specified directory
        """
        gen = self.search_engine._get_image(keywords)
        img_count = 0
        img_filename_ind = 1
        for url in gen:
            if url == SearchEngine.END_SIGNAL:  # no more search results
                return

            file_type = url[url.rfind('.') + 1:]
            file_type = (file_type[:4]) if len(file_type) > 4 else file_type
            if file_type.lower() not in ["jpeg", "jfif", "exif", "tiff", "gif", "bmp", "png", "webp", "jpg"]:
                file_type = "jpg"

            filename = f'{prefix}{img_filename_ind}{postfix}.{file_type}'
            while os.path.exists(os.path.join(directory, filename)):
                img_filename_ind += 1
                filename = f'{prefix}{img_filename_ind}{postfix}.{file_type}'

            if download_file(directory, filename, url, min_width, min_height):
                url_log = ''
                if url.startswith('http'):
                    url_log = f'from {url}'
                logger.info(f"Downloaded file: {filename}{url_log}")
                img_count += 1

            if img_count >= limit:
                return




"""
Template for extracting image links on bing
"""
from image_mint import SearchEngine
from selenium.webdriver.common.by import By
from urllib.parse import quote


class Bing(SearchEngine):
    """Implementation of Bing search engine"""
    def __init__(self, *args):
        super().__init__(*args)
        self.img_class_ref = 'mimg'
        self.has_next_result_btn = True
        self.must_type_search = False

    def _get_search_url(self, search):
        search = quote(search.encode('utf-8'))
        return f"https://www.bing.com/images/search?q={search}&form=IRBPRS&first=1&tsc=ImageHoverTitle"

    def _get_next_btn(self):
        return self.driver.find_element(By.CLASS_NAME, "btn_seemore")




import unittest
from image_mint import Scraper, SearchEngine


class MockSearch(SearchEngine):
    def __init__(self, path):
        pass

    def _get_image(self, keywords):
        yield SearchEngine.END_SIGNAL


class TestFeatures(unittest.TestCase):
    def test_download(self):
        scraper = Scraper(MockSearch('path'))
        scraper.download("test", "dir")


if __name__ == '__main__':
    unittest.main()



__all__ = ["bing", "dogpile", "duckduckgo", "yahoo", "google"]

from .bing import Bing
from .dogpile import DogPile
from .duckduckgo import DuckDuckGo
from .google import Google
from .yahoo import Yahoo



"""
Template for extracting image links on dogpile
"""
from image_mint import SearchEngine
from selenium.webdriver.common.by import By


class DogPile(SearchEngine):
    """Implementation of DogPile search engine"""
    def __init__(self, *args):
        super().__init__(*args)
        self.img_class_ref = None
        self.has_next_result_btn = True
        self.must_type_search = True

    def _get_search_url(self, search):
        return "https://www.dogpile.com/?qc=images"

    def _get_next_btn(self):
        return self.driver.find_element(By.CLASS_NAME, "pagination__num--next")

    def _get_search_field(self):
        return self.driver.find_element(By.NAME, 'q')



"""
Template for extracting image links on duckduckgo
"""
from image_mint import SearchEngine
from urllib.parse import quote


class DuckDuckGo(SearchEngine):
    """Implementation of DuckDuckGo search engine"""
    def __init__(self, *args):
        super().__init__(*args)
        self.img_class_ref = 'tile--img__img'
        self.has_next_result_btn = False

    def _get_search_url(self, search):
        search = quote(search.encode('utf-8'))
        return f"https://duckduckgo.com/?q={search}&t=ha&va=j&iax=images&ia=images"



"""
Template for extracting image links on google
"""
from urllib.parse import quote
from image_mint import SearchEngine
from selenium.webdriver.common.by import By


class Google(SearchEngine):
    """Implementation of Google search engine"""
    def __init__(self, *args):
        super().__init__(*args)
        self.img_class_ref = 'rg_i'
        self.has_next_result_btn = True
        self.must_type_search = False

    def _get_search_url(self, search):
        search = quote(search.encode('utf-8'))
        return f"https://www.google.com/search?q={search}&espv=2&biw=1366&bih=667&site=webhp&source=lnms&tbm=isch&sa=X&ei=XosDVaCXD8TasATItgE&ved=0CAcQ_AUoAg"

    def _get_next_btn(self):
        return self.driver.find_element(By.XPATH, "//input[@value='Show more results']")



"""
image_mint
-----------
image_mint is a python package for downloading images from search engines
`````
* Source
  https://github.com/kouroshparsa/image_mint
"""
from setuptools import setup, find_packages
version = '1.0.3'
long_description = ''
with open("README.md", "r") as fh:
    long_description = fh.read()


setup(
    name="image_mint",
    version=version,
    author="Kourosh Parsa",
    author_email="kouroshtheking@gmail.com",
    description='image_mint is a python package for searching for and downloading images',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/kouroshparsa/image_mint",
    keywords=['google', 'bing', 'yahoo', 'images', 'scraping', 'search'],
    packages=find_packages(),
    install_requires=[
        'selenium',
        'requests',
        'argparse',
        'Pillow'
    ],
    classifiers=(
        "Environment :: Console",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ),
    entry_points={
        'console_scripts': ['image_mint = image_mint.console:main'],
    }
)


"""
Used for running in the terminal
"""
import os
import argparse
import image_mint.engines as engines
from image_mint import Scraper


def main():
    parser = argparse.ArgumentParser(
        description="Scrape images from the internet.")
    parser.add_argument("search", help="Search keywords")
    parser.add_argument("-e", "--engine", required=False, default="Google",
                        choices=['Google', 'Bing', 'DogPile', 'Yahoo', 'DuckDuckGo'],
                        help="Which search engine should be used? (DogPile/Bing/Google/DuckDuckGo/Yahoo)")
    parser.add_argument("-l", "--limit", help="Number of images to download.", type=int,
                        default=100, required=False)
    parser.add_argument("--prefix", help="Filename prefix", default='', required=False)
    parser.add_argument("--postfix", help="Filename postfix", default='', required=False)
    parser.add_argument("-d", "--directory", help="Directory where to download the files",
                        default=None, required=True)
    parser.add_argument("-c", "--chrome_driver",
                        help="Path to Chrome driver executable. The driver can be Downloaded ",
                        default=None, required=True)
    parser.add_argument("-mw", "--min_width", help="Minimum acceptable image width",
                        type=int, default=None,
                        required=False)
    parser.add_argument("-mh", "--min_height", help="Minimum acceptable image height",
                        type=int, default=None,
                        required=False)
    args = parser.parse_args()
    engine = getattr(engines, args.engine)(args.chrome_driver)
    if not os.path.exists(args.directory):
        raise Exception(f'The directory does not exist: {args.directory}')
    scraper = Scraper(engine)
    scraper.download(args.search, args.directory, limit=args.limit, min_width=args.min_width,
                     min_height=args.min_height, prefix=args.prefix, postfix=args.postfix)


if __name__ == "__main__":
    main()



"""
Template for extracting image links on yahoo
"""
from image_mint import SearchEngine
from selenium.webdriver.common.by import By
from urllib.parse import quote


class Yahoo(SearchEngine):
    """Implementation of Yahoo search engine"""
    def __init__(self, *args):
        super().__init__(*args)
        self.img_class_ref = 'tile--rg_i'
        self.has_next_result_btn = True
        self.must_type_search = False

    def _get_search_url(self, search):
        search = quote(search.encode('utf-8'))
        return f"https://ca.images.search.yahoo.com/search/images?p={search}"

    def _get_next_btn(self):
        return self.driver.find_element(By.XPATH, "//button[@value='more-res']")
